#include <cmath>
#include <iostream>
#include <sstream>
#include <iomanip>

using namespace std;

// M_E and M_PI are not part of ISO C++
#ifndef M_E
#define M_E             2.7182818284590452354
#endif
#ifndef M_PI
#define M_PI            3.14159265358979323846
#endif

typedef double Real;

#define BLOCK_SIZE 16

typedef Real Matrix[BLOCK_SIZE+2][BLOCK_SIZE+2];

__global__ void jacobi(Matrix A, int nofiterations) {
   int i = threadIdx.x + 1;
   int j = threadIdx.y + 1;
   for (int it = 0; it < nofiterations; ++it) {
      Real Aij = 0.25 * (A[i-1][j] + A[i][j-1] + A[i][j+1] + A[i+1][j]);
      __syncthreads();
      A[i][j] = Aij;
      __syncthreads();
   }
}

void initialize_A(Matrix A) {
   const unsigned int N = BLOCK_SIZE + 1;
   const static Real E_POWER_MINUS_PI = powf(M_E, -M_PI);
   for (int i = 0; i <= N; ++i) {
      for (int j = 0; j <= N; ++j) {
	 if (j == 0) {
	    A[i][j] = sinf(M_PI * ((Real)i/N));
	 } else if (j == N) {
	    A[i][j] = sinf(M_PI * ((Real)i/N)) * E_POWER_MINUS_PI;
	 } else {
	    A[i][j] = 0;
	 }
      }
   }
}

char* cmdname;
void usage() {
   cerr << "Usage: " << cmdname << " [# iterations]" << endl;
   exit(1);
}

int main(int argc, char** argv) {
   int nofiterations = 10000;

   cmdname = *argv++; --argc;
   if (argc > 0) {
      istringstream arg(*argv++); --argc;
      if (!(arg >> nofiterations) || nofiterations <= 0) usage();
   }
   if (argc > 0) usage();

   Matrix A; initialize_A(A);
   Matrix* cuda_A; cudaMalloc((void**)&cuda_A, sizeof(Matrix));
   cudaMemcpy(cuda_A, A, sizeof(Matrix), cudaMemcpyHostToDevice);
   dim3 block(BLOCK_SIZE, BLOCK_SIZE);
   jacobi<<<1, block>>>(*cuda_A, nofiterations);
   cudaMemcpy(A, cuda_A, sizeof(Matrix), cudaMemcpyDeviceToHost);
   cudaFree(cuda_A);

   cout << BLOCK_SIZE + 2 << endl;
   for (int i = 0; i < BLOCK_SIZE + 2; ++i) {
      for (int j = 0; j < BLOCK_SIZE + 2; ++j) {
	 cout << " " << A[i][j];
      }
      cout << endl;
   }
}
