#include <cassert>
#include <iostream>
#include <fstream>
#include <cstdlib>
#include <iomanip>
#include <sys/time.h>

using namespace std;
typedef double Real;

#define BLOCK_SIZE 16

struct Matrix {
   unsigned int N;
   Real* data;
   bool cuda_allocated;
   Real* cuda_data;

   Matrix() :
      N(0), data(0), cuda_allocated(false), cuda_data(0) {
   }
   ~Matrix() {
      if (data) delete data;
      if (cuda_allocated) release_cuda_data();
   }

   bool copy_to_gpu() {
      if (!cuda_allocated) {
	 if (!allocate_cuda_data()) return false;
      }
      return cudaMemcpy(cuda_data, data, N * N * sizeof(Real),
	       cudaMemcpyHostToDevice) == cudaSuccess;
   }

   bool copy_from_gpu() {
      assert(cuda_allocated);
      return cudaMemcpy(data, cuda_data, N * N * sizeof(Real),
	       cudaMemcpyDeviceToHost) == cudaSuccess;
   }

   bool allocate_cuda_data() {
      if (cuda_allocated) return true;
      Real* cudap;
      if (cudaMalloc((void**)&cudap, N * N * sizeof(Real)) != cudaSuccess) {
	 return false;
      }
      cuda_data = cudap;
      cuda_allocated = true;
      return true;
   }

   void release_cuda_data() {
      if (cuda_data) {
	 cudaFree(cuda_data);
	 cuda_data = 0;
      }
   }

   bool resize(unsigned int N_) {
      if (N == N_) return true;
      Real* rp = new Real[N_ * N_];
      if (!rp) return false;
      if (data) delete data;
      release_cuda_data();
      data = rp; N = N_;
      return true;
   }

   Real& operator()(unsigned int i, unsigned int j) {
      return data[i*N + j];
   }

   const Real& operator()(unsigned int i, unsigned int j) const {
      return data[i*N + j];
   }
};

// input operator for square matrices that expects
// an input sequence consisting of N, followed by N*N
// elements of the matrix in row-major order
istream& operator>>(istream& in, Matrix& matrix) {
   unsigned int N;
   if (!(in >> N)) return in;
   if (N == 0 || !matrix.resize(N)) {
      in.setstate(ios::failbit); return in;
   }
   matrix.N = N;
   for (int i = 0; i < N; ++i) {
      for (int j = 0; j < N; ++j) {
	 if (!(in >> matrix(i,j))) return in;
      }
   }
   return in;
}

ostream& operator<<(ostream& out, Matrix& matrix) {
   unsigned int N = matrix.N;
   for (int i = 0; i < N; ++i) {
      for (int j = 0; j < N; ++j) {
	 if (!(out << " " << matrix(i, j))) return out;
      }
      cout << endl;
   }
   return out;
}

// fetch a matrix from an input file
bool read_matrix(char* filename, Matrix& matrix) {
   fstream in(filename);
   return in >> matrix;
}

#define ELEMENT(m,i,j) ((m)[(i) * stride + (j)])

__global__ void mmm(Real* a, Real* b, Real* c) {
   unsigned int stride = gridDim.y * BLOCK_SIZE;
   unsigned int row = blockIdx.y * BLOCK_SIZE + threadIdx.y;
   unsigned int col = blockIdx.x * BLOCK_SIZE + threadIdx.x;

   Real sum = 0;
   for (int k = 0; k < BLOCK_SIZE * gridDim.y; ++k) {
      sum += ELEMENT(a, row, k) * ELEMENT(b, k, col);
   }
   ELEMENT(c, row, col) = sum;
}

char* cmdname;
void usage() {
   cerr << "Usage: " << cmdname << " A_file B_file" << endl;
   exit(1);
}

int main(int argc, char** argv) {
   cmdname = *argv++; --argc;
   if (argc != 2) usage();
   Matrix A; if (!read_matrix(*argv++, A)) usage(); --argc;
   Matrix B; if (!read_matrix(*argv++, B)) usage(); --argc;
   cout << "A = " << endl << A << endl;
   cout << "B = " << endl << B << endl;
   if (A.N != B.N) {
      cerr << cmdname << ": sizes of the matrices do not match" << endl;
      exit(1);
   }
   if (A.N % BLOCK_SIZE) {
      cerr << cmdname << ": size of matrices is not a multiply of "
	 << BLOCK_SIZE << endl;
      exit(1);
   }

   A.copy_to_gpu();
   B.copy_to_gpu();
   Matrix C; C.resize(A.N); C.allocate_cuda_data();
   dim3 block(BLOCK_SIZE, BLOCK_SIZE);
   dim3 grid(A.N / BLOCK_SIZE, A.N / BLOCK_SIZE);

   struct timeval tbuf1; gettimeofday(&tbuf1, 0);

   mmm<<<grid, block>>>(A.cuda_data, B.cuda_data, C.cuda_data);

   struct timeval tbuf2; gettimeofday(&tbuf2, 0);
   long int usec = (tbuf2.tv_sec - tbuf1.tv_sec) * 1000000 +
                   (tbuf2.tv_usec - tbuf1.tv_usec);
   double timeInMillisecs = (double) usec / (double) 1000;
   cerr << "GPU time in ms: " << timeInMillisecs << endl;

   C.copy_from_gpu();
   cout << "C = " << endl << setprecision(14) << C << endl;
}
