#include <cmath>
#include <iostream>
#include <sstream>
#include <iomanip>

using namespace std;

typedef double Real;

/* return unique id within a block */
__device__ int get_threadid() {
   return threadIdx.z * blockDim.x * blockDim.y +
	  threadIdx.y * blockDim.x +
	  threadIdx.x;
}

/* return block id a thread is associated to */
__device__ int get_blockid() {
   return blockIdx.x + blockIdx.y * gridDim.x;
}

/* return number of threads per block */
__device__ int get_nofthreads_per_block() {
   return blockDim.x * blockDim.y * blockDim.z;
}

/* return number of blocks */
__device__ int get_nofblocks() {
   return gridDim.x * gridDim.y;
}

/* return total number of threads */
__device__ int get_nofthreads() {
   return get_nofthreads_per_block() * get_nofblocks();
}

/* return id which is unique throughout all threads */
__device__ int get_id() {
   return get_blockid() * blockDim.x * blockDim.y * blockDim.z +
          get_threadid();
}

int max_threads_per_block() {
   int device; cudaGetDevice(&device);
   struct cudaDeviceProp device_prop;
   cudaGetDeviceProperties(&device_prop, device);
   return device_prop.maxThreadsPerBlock;
}

// to be integrated function
__device__ Real f(Real x) {
   return 4 / (1 + x*x);
}

// numerical integration according to the Simpson rule
// for f over the interval [a,b]
__global__ void simpson(Real a, Real b, Real* sums) {
   const int N = get_nofthreads();
   const int i = get_id();
   Real xleft = a + (b - a) / N * i;
   Real xright = xleft + (b - a) / N;
   Real xmid = (xleft + xright) / 2;
   sums[i] = (xright - xleft) / 6 * (f(xleft) + 4 * f(xmid) + f(xright));
}

char* cmdname;
void usage() {
   cerr << "Usage: " << cmdname << " [# intervals]" << endl;
   exit(1);
}

int main(int argc, char** argv) {
   const Real a = 0;
   const Real b = 1;
   int N = 8192;

   cmdname = *argv++; --argc;
   if (argc > 0) {
      istringstream arg(*argv++); --argc;
      if (!(arg >> N) || N <= 0) usage();
   }
   if (argc > 0) usage();

   int blocksize = max_threads_per_block();
   if (blocksize > N) {
      blocksize = N;
   } else {
      if (N % blocksize != 0) {
	 cerr << cmdname << ": please select a multiple of "
	    << blocksize << endl;
	 exit(1);
      }
   }
   int nof_blocks = N / blocksize;

   dim3 blockdim(blocksize, 1, 1);
   dim3 griddim(nof_blocks, 1);

   Real sums[N];
   Real* cuda_sums; cudaMalloc((void**)&cuda_sums, N * sizeof(Real));
   simpson<<<griddim, blockdim>>>(a, b, cuda_sums);
   cudaMemcpy(sums, cuda_sums, N * sizeof(Real), cudaMemcpyDeviceToHost);
   cudaFree(cuda_sums);
   double sum = 0;
   for (int i = 0; i < N; ++i) {
      sum += sums[i];
   }
   cout << setprecision(14) << sum << endl;
   cout << setprecision(14) << M_PI << endl;
}
